from ayugespidertools.scraper.http.request import AiohttpRequest
from ayugespidertools.scraper.http.request.form import AiohttpFormRequest

__all__ = [
    "AiohttpRequest",
    "AiohttpFormRequest",
]
